<?php
/*
    $Id: upsxml.php,v 1.73 2018-07-12 $

    Original copyright (c) 2003 Torin Walker
    This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.
    You should have received a copy of the GNU General Public License along with this program;
    If not, you may obtain one by writing to and requesting one from:
    The Free Software Foundation, Inc.,
    59 Temple Place, Suite 330,
    Boston, MA 02111-1307 USA

    Written by Torin Walker.
    Some code/style borrowed from both Fritz Clapp's UPS Choice 1.7 Module,
    and Kelvin, Kenneth, and Tom St.Croix's Canada Post 3.1 Module.
    Insurance support by Joe McFrederick
*/

// Incorporate the XML conversion library 
  require_once (DIR_WS_CLASSES . 'xml.php');

class upsxml {
    var $code, $title, $description, $icon, $enabled, $types, $boxcount;

    function __construct() {
			$this->upsxml();
		}
		
    //***************
    function upsxml() {
        global $order, $packing;
        $this->code = 'upsxml';
        $this->title = MODULE_SHIPPING_UPSXML_RATES_TEXT_TITLE;
        $this->description = MODULE_SHIPPING_UPSXML_RATES_TEXT_DESCRIPTION;
        $this->enabled = (defined('MODULE_SHIPPING_UPSXML_RATES_STATUS') && (MODULE_SHIPPING_UPSXML_RATES_STATUS == 'True') ? true : false);
				if (!$this->enabled) {$this->sort_order = 0; return;} // if module is not enabled no further settings are needed
        $this->sort_order = MODULE_SHIPPING_UPSXML_RATES_SORT_ORDER;
        $this->icon = DIR_WS_ICONS . 'shipping_ups.gif';
        $this->tax_class = MODULE_SHIPPING_UPSXML_RATES_TAX_CLASS;
        $this->access_key = MODULE_SHIPPING_UPSXML_RATES_ACCESS_KEY;
        $this->access_username = MODULE_SHIPPING_UPSXML_RATES_USERNAME;
        $this->access_password = MODULE_SHIPPING_UPSXML_RATES_PASSWORD;
        $this->access_account_number = MODULE_SHIPPING_UPSXML_RATES_UPS_ACCOUNT_NUMBER;
        // BLM 2-14-08 MANUAL NEGOTIATED RATE
		    // set the rate
		    $this->manual_negotiated_rate = intval(MODULE_SHIPPING_UPSXML_RATES_MANUAL_NEGOTIATED_RATE);
        $this->use_negotiated_rates = MODULE_SHIPPING_UPSXML_RATES_USE_NEGOTIATED_RATES;
        $this->origin = MODULE_SHIPPING_UPSXML_RATES_ORIGIN;
        $this->origin_city = MODULE_SHIPPING_UPSXML_RATES_CITY;
        $this->origin_stateprov = MODULE_SHIPPING_UPSXML_RATES_STATEPROV;
        $this->origin_country = MODULE_SHIPPING_UPSXML_RATES_COUNTRY;
        $this->origin_postalcode = MODULE_SHIPPING_UPSXML_RATES_POSTALCODE;
        $this->pickup_method = MODULE_SHIPPING_UPSXML_RATES_PICKUP_METHOD;
        $this->package_type = MODULE_SHIPPING_UPSXML_RATES_PACKAGE_TYPE;
// the variables for unit weight, unit length, and dimensions support were moved to
// shop admin -> Configuration -> Shipping/Packaging in
// version 1.3.0. Run the configuration_shipping.sql to add these to your configuration
        if (defined('SHIPPING_UNIT_WEIGHT')) {
          $this->unit_weight = SHIPPING_UNIT_WEIGHT;
        } else {
// for those who will undoubtedly forget or not know how to run the configuration_shipping.sql
// we will set the default to pounds (LBS) and inches (IN)
          $this->unit_weight = 'LBS';
        }
        if (defined('SHIPPING_UNIT_LENGTH')) {
          $this->unit_length = SHIPPING_UNIT_LENGTH;
        } else {
          $this->unit_length = 'IN';
        }
        if (defined('SHIPPING_DIMENSIONS_SUPPORT') && SHIPPING_DIMENSIONS_SUPPORT == 'Ready-to-ship only') {
          $this->dimensions_support = 1;
        } elseif (defined('SHIPPING_DIMENSIONS_SUPPORT') && SHIPPING_DIMENSIONS_SUPPORT == 'With product dimensions') {
          $this->dimensions_support = 2;
        } else {
          $this->dimensions_support = 0;
        }
        $this->email_errors = ((MODULE_SHIPPING_UPSXML_EMAIL_ERRORS == 'Yes') ? true : false);
        $this->handling_type = MODULE_SHIPPING_UPSXML_HANDLING_TYPE;
        $this->handling_fee = MODULE_SHIPPING_UPSXML_RATES_HANDLING;
        $this->handling_charged = MODULE_SHIPPING_UPSXML_HANDLING_CHARGED;
        $this->quote_type = MODULE_SHIPPING_UPSXML_RATES_QUOTE_TYPE;
        $this->customer_classification = MODULE_SHIPPING_UPSXML_RATES_CUSTOMER_CLASSIFICATION_CODE;
        $this->protocol = 'https';
        $this->host = ((MODULE_SHIPPING_UPSXML_RATES_MODE == 'Test') ? 'wwwcie.ups.com' : 'onlinetools.ups.com');
        $this->port = '443';
        $this->path = '/ups.app/xml/Rate';
        $this->transitpath = '/ups.app/xml/TimeInTransit';
        $this->version = 'UPSXML Rate 1.0001';
        $this->timeout = '60';
        $this->xpci_version = '1.0001';
        $this->api_subversion = '1801';
        $this->items_qty = 0;
        $this->timeintransit = '0';
        $this->timeInTransitView = MODULE_SHIPPING_UPSXML_RATES_TIME_IN_TRANSIT_VIEW;
        $this->weight_for_timeintransit = '0';
        $this->dayname = array('Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday');
        $tmp = explode(', ', MODULE_SHIPPING_UPSXML_STORE_DAYS);
        $this->store_days = array();
        foreach ($tmp as $t)
          // make sure only valid entries are stored
          if (in_array($t, $this->dayname)) $this->store_days[] = $t;
        // if no valid entries set one day to prevent endless loop
        if (empty($this->store_days)) $this->store_days[] = 'Friday';
        $tmp = explode(', ', MODULE_SHIPPING_UPSXML_UPS_PICKUP_DAYS);
        $this->ups_days = array();
        foreach ($tmp as $t)
          // make sure only valid entries are stored
          if (in_array($t, $this->dayname)) $this->ups_days[] = $t;
        // if no valid entries set one day to prevent endless loop
        if (empty($this->ups_days)) $this->ups_days[] = 'Friday';
        $this->ups_holidays = array();
        $this_year = date('Y');
        if (substr(MODULE_SHIPPING_UPSXML_UPS_HOLIDAYS, 0, 4) == $this_year) {
          $tmp = explode(',', substr(MODULE_SHIPPING_UPSXML_UPS_HOLIDAYS, 5));
          foreach ($tmp as $t)
            if ((strlen($t) == 4) && checkdate(substr($t, 0, 2), substr($t, 2, 2), $this_year)) $this->ups_holidays[] = $t;
        }
        if (empty($this->ups_holidays)) $this->_update_ups_holidays($this_year);

        // insurance addition
        $this->insure_package = (MODULE_SHIPPING_UPSXML_INSURE == 'True');
        // end insurance addition
				$this->pkgvalue = 0; // make sure this is set
        if (is_object($order)) $this->pkgvalue = ceil($order->info['subtotal']); // is divided by number of boxes later
        $this->all_adult_sig = $this->all_sig = false;
        if (MODULE_SHIPPING_UPSXML_REQSIG == 'Shipment Value') {
          if ($this->pkgvalue >= MODULE_SHIPPING_UPSXML_ADULTSIG_THRESHOLD) {
            $this->all_adult_sig = true;
          } elseif ($this->pkgvalue >= MODULE_SHIPPING_UPSXML_SIG_THRESHOLD) {
            $this->all_sig = true;
          }
        }
			  $this->sigreq = $this->sig_level = ''; // no signature requirement
        // to enable logging, create an empty "upsxml.log" file at the location you set below, give it write permissions (777) and uncomment the next line
//        $this->logfile = '/srv/www/htdocs/catalog/includes/modules/shipping/upsxml.log';
        // to enable logging of just the errors, do as above but call the file upsxml_error.log
//        $this->ups_error_file = '/srv/www/htdocs/catalog/includes/modules/shipping/upsxml_error.log';
        // when cURL is not compiled into PHP (Windows users, some Linux users)
        // you can set the next variable to "1" and then exec(curl -d $xmlRequest, $xmlResponse)
        // will be used
        $this->use_exec = '0';
				$this->disallowed_methods = explode(', ',MODULE_SHIPPING_UPSXML_TYPES);
				$this->disallowed_methods[] = '94'; // not SurePost BPM since module can't know if all items are bound printed matter
				$this->disallowed_methods[] = '95'; // not SurePost Media since module can't know if all items are media
				if (MODULE_SHIPPING_UPSXML_SUREPOST == 'No') {
				  $this->disallowed_methods[] = '92'; // not SurePost < 1 pound
				  $this->disallowed_methods[] = '93'; // not SurePost 1 pound or greater
				}

        if (($this->enabled == true) && ((int)MODULE_SHIPPING_UPSXML_RATES_ZONE > 0)) {
            $check_flag = false;
            $check_query = tep_db_query("select zone_id from " . TABLE_ZONES_TO_GEO_ZONES . " where geo_zone_id = '" . MODULE_SHIPPING_UPSXML_RATES_ZONE . "' and zone_country_id = '" . $order->delivery['country']['id'] . "' order by zone_id");
            while ($check = tep_db_fetch_array($check_query)) {
                if ($check['zone_id'] < 1) {
                    $check_flag = true;
                    break;
                } elseif ($check['zone_id'] == $order->delivery['zone_id']) {
                    $check_flag = true;
                    break;
                }
            }
            if ($check_flag == false) {
                $this->enabled = false;
            }
        }

        // Available pickup types - set in admin
        $this->pickup_methods = array(
            'Daily Pickup' => '01',
            'Customer Counter' => '03',
            'One Time Pickup' => '06',
            'Letter Center' => '19',
            'Air Service Center' => '20'
        );

        // Available package types
        $this->package_types = array(
            'UPS Letter' => '01',
            'Package' => '02',
            'UPS Tube' => '03',
            'UPS Pak' => '04',
            'UPS Express Box' => '21',
            'UPS 25kg Box' => '24',
            'UPS 10kg Box' => '25'
        );
    } // end function upsxml

    // class methods
    function quote($method = '') {
      global $HTTP_POST_VARS, $order, $shipping_weight, $shipping_num_boxes, $total_weight, $boxcount, $cart, $packing;
    	$upsxml_text_package = array(true => MODULE_SHIPPING_UPSXML_TEXT_PKGS, false => MODULE_SHIPPING_UPSXML_TEXT_PKG);
    	$upsxml_text_weight = array('LBS' => MODULE_SHIPPING_UPSXML_TEXT_LBS, 'KGS' => MODULE_SHIPPING_UPSXML_TEXT_KGS, 'OZS' => MODULE_SHIPPING_UPSXML_TEXT_OZ);
      $this->_upsOrigin(MODULE_SHIPPING_UPSXML_RATES_CITY, MODULE_SHIPPING_UPSXML_RATES_STATEPROV, MODULE_SHIPPING_UPSXML_RATES_COUNTRY, MODULE_SHIPPING_UPSXML_RATES_POSTALCODE);
      $this->_upsDest($order->delivery);
			$this->_customer_name = $order->customer['firstname'] . ' ' . $order->customer['lastname'];
			if (MODULE_SHIPPING_UPSXML_REQSIG != 'No') {
				$this->sig_level = 'S'; // signature must be requested at shipment level unless package level request is determined below
				if ($this->_upsOriginCountryCode == 'US' && ($this->_upsDestCountryCode == 'US' || $this->_upsDestCountryCode == 'PR')) $this->sig_level = 'P';
				if ($this->_upsOriginCountryCode == 'PR' && ($this->_upsDestCountryCode == 'US' || $this->_upsDestCountryCode == 'PR')) $this->sig_level = 'P';
				if ($this->_upsOriginCountryCode == 'CA' && $this->_upsDestCountryCode == 'CA') $this->sig_level = 'P';
			}
        
// the check on $packing being an object will puzzle people who do things wrong (no changes when 
// you enable dimensional support without changing checkout_shipping.php) but better be safe
      if ($this->dimensions_support > 0 && is_object($packing)) {
        $boxValue = 0;
        $this->totalWeight = $packing->getTotalWeight();
        $boxesToShip = $packing->getPackedBoxes();
				$this->NumPieces = count($boxesToShip);
        for ($i = 0; $i < count($boxesToShip); $i++) {
          $this->_addItem($boxesToShip[$i]['item_length'], $boxesToShip[$i]['item_width'], $boxesToShip[$i]['item_height'], $boxesToShip[$i]['item_weight'], $boxesToShip[$i]['item_price']);
        } // end for ($i = 0; $i < count($boxesToShip); $i++)
      } else {
        // The old method. Let osCommerce tell us how many boxes, plus the weight of each (or total? - might be sw/num boxes)
			  $this->totalWeight = $total_weight;
				$this->NumPieces = $shipping_num_boxes;
        $this->items_qty = 0; //reset quantities
        // $this->pkgvalue has been set as order subtotal around line 108, it will cause overcharging
        // of insurance if not divided by the number of boxes
        for ($i = 0; $i < $shipping_num_boxes; $i++) {
          $this->_addItem(0, 0, 0, $shipping_weight, number_format(($this->pkgvalue/$shipping_num_boxes), 2, '.', ''));
        }
      }

      $upsQuote = $this->_upsGetQuote();
			// debugging line
			//print_r($upsQuote);
      if ((is_array($upsQuote)) && (sizeof($upsQuote) > 0)) {
        uasort($upsQuote, array($this, "rate_sort_func")); // uasort maintains keys and usort doesn't
				$title = $this->title;
        if (MODULE_SHIPPING_UPSXML_WEIGHT == 'True') {
          $title .= ' (' . $this->NumPieces . ' ' . $upsxml_text_package[$this->NumPieces > 1] . ', ' . $this->totalWeight . ' ' . $upsxml_text_weight[$this->unit_weight] . MODULE_SHIPPING_UPSXML_TEXT_TOTAL . ')';
        } // end else/if if (defined('MODULE_SHIPPING_UPSXML_WEIGHT')
        $this->quotes = array('id' => $this->code, 'module' => $title);
        $methods = array();
        foreach ($upsQuote as $code => $svc) {
          $basetype = $svc['method'];
          if ($method == '' || $method == $basetype) {
            $method_name = $basetype;
            if (MODULE_SHIPPING_UPSXML_WEIGHT == 'True')
              $method_name .= ' (' . MODULE_SHIPPING_UPSXML_TEXT_BILLED_WEIGHT . $svc['billed_weight'] . ' ' . $upsxml_text_weight[$svc['weight_code']] . ')';
						if (MODULE_SHIPPING_UPSXML_INSURE == 'True') {
							$method_name .= '<br>---';
							if (in_array($code, array('92', '93', '94', '95'))) { // SurePost cannot be insured
								$method_name .= MODULE_SHIPPING_UPSXML_TEXT_NOTINSURED;
							} else {
								$method_name .= MODULE_SHIPPING_UPSXML_TEXT_INSURED;
							}
						}
						if ($this->sigreq != '') {
							$method_name .= '<br>---';
							if ($this->sigreq == 'a') { // adult signature required
								$method_name .= MODULE_SHIPPING_UPSXML_TEXT_ASIG;
							} else { // signature required
								$method_name .= MODULE_SHIPPING_UPSXML_TEXT_SIG;
							}
						}
            if ($this->timeInTransitView == "Raw") {
							$method_name .= '<br>---' . MODULE_SHIPPING_UPSXML_TEXT_EDD . $svc['edd'];
            } elseif ($this->timeInTransitView == "Detailed") {
							if (in_array($code, array('92', '93', '94', '95'))) { 
								$method_name .= '<br>---' . MODULE_SHIPPING_UPSXML_TEXT_EDD . MODULE_SHIPPING_UPSXML_TEXT_SUREPOST_DEL;
							} else {
							  $method_name .= '<br>---' . MODULE_SHIPPING_UPSXML_TEXT_EDD . tep_date_long($svc['edd']);
							}
            }                    
            if ($this->handling_type == "Percentage") {
              $cost = (($this->handling_fee * $svc['charge'])/100) + $svc['charge'];
            } else {
              if ($this->handling_charged == 'Per Shipment') {
                $cost = $this->handling_fee + $svc['charge'];
              } else {
                $cost = ($this->handling_fee * $this->NumPieces) + $svc['charge'];
              }
            }
					  $methods[] = array('id' => $basetype, 'title' => $method_name, 'cost' => $cost);
          }
        }
        if ($this->tax_class > 0) {
          $this->quotes['tax'] = tep_get_tax_rate($this->tax_class, $order->delivery['country']['id'], $order->delivery['zone_id']);
        }
        $this->quotes['methods'] = $methods;
      } else {
          if ($upsQuote === false) {
            $errmsg = MODULE_SHIPPING_UPSXML_RATES_TEXT_UNKNOWN_ERROR;
          } elseif (empty($upsQuote)) {
            $errmsg = MODULE_SHIPPING_UPSXML_RATES_TEXT_RATES_ERROR;
          } else {
            $errmsg = $upsQuote;
					}
          $errmsg .= '<br>' . MODULE_SHIPPING_UPSXML_RATES_TEXT_IF_YOU_PREFER . ' ' . STORE_NAME.' via <a href="mailto:'.STORE_OWNER_EMAIL_ADDRESS.'"><u>' . MODULE_SHIPPING_UPSXML_RATES_TEXT_EMAIL . '</u></a>.';
          $this->quotes = array('module' => $this->title, 'error' => $errmsg);
        }
        if (tep_not_null($this->icon)) {
            $this->quotes['icon'] = tep_image($this->icon, $this->title);
        }
        return $this->quotes;
    }

    //**************
    function check() {
        if (!isset($this->_check)) {
            $check_query = tep_db_query("select configuration_value from " . TABLE_CONFIGURATION . " where configuration_key = 'MODULE_SHIPPING_UPSXML_RATES_STATUS'");
            $this->_check = tep_db_num_rows($check_query);
        }
        return $this->_check;
    }

    //**************
    function install() {
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Enable UPS Shipping', 'MODULE_SHIPPING_UPSXML_RATES_STATUS', 'True', 'Do you want to offer UPS shipping?', '6', '0', 'tep_cfg_select_option(array(\'True\', \'False\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('UPS Rates Access Key', 'MODULE_SHIPPING_UPSXML_RATES_ACCESS_KEY', '', 'Enter the XML rates access key assigned to you by UPS.', '6', '1', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('UPS Rates Username', 'MODULE_SHIPPING_UPSXML_RATES_USERNAME', '', 'Enter your UPS Services account username.', '6', '2', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('UPS Rates Password', 'MODULE_SHIPPING_UPSXML_RATES_PASSWORD', '', 'Enter your UPS Services account password.', '6', '3', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Pickup Method', 'MODULE_SHIPPING_UPSXML_RATES_PICKUP_METHOD', 'Daily Pickup', 'How do you give packages to UPS (only used when origin is US)?', '6', '4', 'tep_cfg_select_option(array(\'Daily Pickup\', \'Customer Counter\', \'One Time Pickup\', \'Letter Center\', \'Air Service Center\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Packaging Type', 'MODULE_SHIPPING_UPSXML_RATES_PACKAGE_TYPE', 'Package', 'What kind of packaging do you use?', '6', '5', 'tep_cfg_select_option(array(\'Package\', \'UPS Letter\', \'UPS Tube\', \'UPS Pak\', \'UPS Express Box\', \'UPS 25kg Box\', \'UPS 10kg box\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Customer Classification Code', 'MODULE_SHIPPING_UPSXML_RATES_CUSTOMER_CLASSIFICATION_CODE', '53', '00 - Rates Associated with Account Number (UPS account number MUST be entered for this), 01 - Daily Rates, 04 - Retail Rates, 05 - Regional Rate, 06 - General List Rate, 53 - Standard List Rates', '6', '6', 'tep_cfg_select_option(array(\'00\', \'01\', \'04\', \'05\', \'06\', \'53\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Shipping Origin', 'MODULE_SHIPPING_UPSXML_RATES_ORIGIN', 'US Origin', 'What origin point should be used (this setting affects only what UPS product names are shown to the user)', '6', '7', 'tep_cfg_select_option(array(\'US Origin\', \'Canada Origin\', \'European Union Origin\', \'Mexico Origin\', \'Poland Origin\', \'Puerto Rico Origin\', \'All other origins\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Shipper Name', 'MODULE_SHIPPING_UPSXML_RATES_SHIPNAME', '', 'Enter the name or company name of the shipper. (optional)', '6', '8', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Shipper Address Line 1', 'MODULE_SHIPPING_UPSXML_RATES_ADDRESS1', '', 'Enter the shipper&rsquo;s street address including name and number.', '6', '8', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Shipper Address Line 2', 'MODULE_SHIPPING_UPSXML_RATES_ADDRESS2', '', 'Enter additional shipper address information, preferably room or floor. (when applicable)', '6', '8', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Shipper Address Line 3', 'MODULE_SHIPPING_UPSXML_RATES_ADDRESS3', '', 'Enter additional shipper address information, preferably department name. (when applicable)', '6', '8', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Origin City', 'MODULE_SHIPPING_UPSXML_RATES_CITY', '', 'Enter the name of the origin city.', '6', '8', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Origin State/Province', 'MODULE_SHIPPING_UPSXML_RATES_STATEPROV', '', 'Enter the two-letter code for your origin state/province.', '6', '9', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Origin Country', 'MODULE_SHIPPING_UPSXML_RATES_COUNTRY', '', 'Enter the two-letter code for your origin country.', '6', '10', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Origin Zip/Postal Code', 'MODULE_SHIPPING_UPSXML_RATES_POSTALCODE', '', 'Enter your origin zip/postal code.', '6', '11', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Test or Production Mode', 'MODULE_SHIPPING_UPSXML_RATES_MODE', 'Test', 'Use this module in Test or Production mode?', '6', '12', 'tep_cfg_select_option(array(\'Test\', \'Production\'), ', now())");
// three configuration options were moved to shop Configuration -> Shipping/Packaging in v1.3.0
// those are (and were renamed to -> ): 
// MODULE_SHIPPING_UPSXML_RATES_UNIT_WEIGHT (LBS/KG) -> SHIPPING_UNIT_WEIGHT
// MODULE_SHIPPING_UPSXML_RATES_UNIT_LENGTH (IN/CM) -> SHIPPING_UNIT_LENGTH
// MODULE_SHIPPING_UPSXML_DIMENSIONS_SUPPORT -> SHIPPING_DIMENSIONS_SUPPORT
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Quote Type', 'MODULE_SHIPPING_UPSXML_RATES_QUOTE_TYPE', 'Residential', 'Quote for Residential or Commercial Delivery', '6', '15', 'tep_cfg_select_option(array(\'Commercial\', \'Residential\'), ', now())");
       // next three keys added to be able to use negotiated rates (available from UPS since about July 2006)
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Negotiated rates', 'MODULE_SHIPPING_UPSXML_RATES_USE_NEGOTIATED_RATES', 'False', 'Do you receive discounted rates from UPS and want to use these for shipping quotes? <strong>Note:</strong>  You need to enter your UPS account number below.', '6', '25', 'tep_cfg_select_option(array(\'True\', \'False\'), ', now())");
         tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Manual Negotiated Rate', 'MODULE_SHIPPING_UPSXML_RATES_MANUAL_NEGOTIATED_RATE', '', 'Enter a negotiated rate manually. <strong>Note:</strong> If \'Negotiated Rates\' above is set to \'True\', This <strong>WILL NOT</strong> be applied. If using this option, set \'Negotiated Rates\' to \'False\'. Usage: \' 57 \' returns 57% of published UPS rate.', '6', '26', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('UPS Account Number', 'MODULE_SHIPPING_UPSXML_RATES_UPS_ACCOUNT_NUMBER', '', 'Enter your UPS Account number when you have and want to use negotiated rates.', '6', '27', now())");
        // added for handling type selection
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Handling Type', 'MODULE_SHIPPING_UPSXML_HANDLING_TYPE', 'Flat Fee', 'Handling type for this shipping method.', '6', '14', 'tep_cfg_select_option(array(\'Flat Fee\', \'Percentage\'), ', now())");
	    	tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Flat Handling Fees Charged', 'MODULE_SHIPPING_UPSXML_HANDLING_CHARGED', 'Per Package', 'Select whether flat fee handling is charged:', '6', '15', 'tep_cfg_select_option(array(\'Per Package\', \'Per Shipment\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Handling Fee', 'MODULE_SHIPPING_UPSXML_RATES_HANDLING', '0', 'Handling fee for this shipping method.', '6', '16', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('UPS Currency Code', 'MODULE_SHIPPING_UPSXML_CURRENCY_CODE', 'USD', 'Enter the 3 letter currency code for your country of origin. United States (USD)', '6', '2', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Enable Insurance', 'MODULE_SHIPPING_UPSXML_INSURE', 'True', 'Do you want to insure packages shipped by UPS?', '6', '22', 'tep_cfg_select_option(array(\'True\', \'False\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, use_function, set_function, date_added) values ('Tax Class', 'MODULE_SHIPPING_UPSXML_RATES_TAX_CLASS', '0', 'Use the following tax class on the shipping fee.', '6', '17', 'tep_get_tax_class_title', 'tep_cfg_pull_down_tax_classes(', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, use_function, set_function, date_added) values ('Shipping Zone', 'MODULE_SHIPPING_UPSXML_RATES_ZONE', '0', 'If a zone is selected, only enable this shipping method for that zone.', '6', '18', 'tep_get_zone_class_title', 'tep_cfg_pull_down_zone_classes(', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Sort order of display.', 'MODULE_SHIPPING_UPSXML_RATES_SORT_ORDER', '0', 'Sort order of display. Lowest is displayed first.', '6', '19', now())");
        // add key for disallowed shipping methods
        tep_db_query("insert into " . TABLE_CONFIGURATION . " ( configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, use_function, set_function, date_added) values ('Disallowed Shipping Methods', 'MODULE_SHIPPING_UPSXML_TYPES', '', 'Select the UPS services <span style=\"color: red; font-weight: bold\">NOT</span> to be offered. Warning: Disallow too many and you may not get any rates returned at all.', '6', '20', 'get_multioption_upsxml',  'upsxml_cfg_select_multioption_indexed(array(\'01\', \'02\', \'03\', \'07\', \'08\', \'11\', \'12\', \'13\', \'14\', \'54\', \'59\', \'65\', \'70\', \'82\', \'83\', \'85\', \'86\', \'96\'), ',  now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Enable UPS SurePost&reg;', 'MODULE_SHIPPING_UPSXML_SUREPOST', 'No', 'Do you want to include quotes for UPS SurePost&reg;? Note: UPS account must have SurePost&reg; enabled and the UPS account number must be entered when configuring this module.', '6', '24', 'tep_cfg_select_option(array(\'Yes\', \'No\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Maximum value by SurePost', 'MODULE_SHIPPING_UPSXML_SUREPOST_MAXVAL', '100', 'What is the maximum value package you are willing to ship by SurePost since it is uninsured?', '6', '19', now())");
        // add key for shipping delay, changed the constant from SHIPPING_DAYS_DELAY in v1.3.0
        tep_db_query("insert into " . TABLE_CONFIGURATION . " ( configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, last_modified, date_added, use_function, set_function) values ('Shipping Delay', 'MODULE_SHIPPING_UPSXML_SHIPPING_DAYS_DELAY', '1', 'How many days from when an order is placed to when you ship it (Decimals are allowed). Arrival date estimations are based on this value.', '6', '21', NULL, now(), NULL, NULL)");
        // add key for enabling email error messages
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Email UPS errors', 'MODULE_SHIPPING_UPSXML_EMAIL_ERRORS', 'Yes', 'Do you want to receive UPS errors by email?', '6', '24', 'tep_cfg_select_option(array(\'Yes\', \'No\'), ', now())");
        // add key for time in transit view type
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Time in Transit View Type', 'MODULE_SHIPPING_UPSXML_RATES_TIME_IN_TRANSIT_VIEW', 'Not', 'If and how the module should display the time in transit to the customer.', '6', '13', 'tep_cfg_select_option(array(\'Not\',\'Raw\', \'Detailed\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " ( configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, use_function, set_function, date_added) values ('Store Processes Orders', 'MODULE_SHIPPING_UPSXML_STORE_DAYS', 'Monday, Tuesday, Wednesday, Thursday, Friday', 'Select the days of the week on which the store processes orders. Used for time in transit delivery date.', '6', '31', NULL,  'upsxml_cfg_select_multioption(array(\'Sunday\', \'Monday\', \'Tuesday\', \'Wednesday\', \'Thursday\', \'Friday\', \'Saturday\'), ',  now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " ( configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, use_function, set_function, date_added) values ('UPS Picks Up', 'MODULE_SHIPPING_UPSXML_UPS_PICKUP_DAYS', 'Monday, Tuesday, Wednesday, Thursday, Friday', 'Select the days of the week on which UPS picks up packages from the store or that the store delivers packages to UPS. Used for time in transit delivery date.', '6', '32', NULL,  'upsxml_cfg_select_multioption(array(\'Monday\', \'Tuesday\', \'Wednesday\', \'Thursday\', \'Friday\'), ',  now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " ( configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, last_modified, date_added, use_function, set_function) values ('UPS Holidays', 'MODULE_SHIPPING_UPSXML_UPS_HOLIDAYS', '2017:1224,1225,1231,0101,0704,1123,1124,0529,0904', 'Format YYYY:MMDD,MMDD...<br />Known UPS holidays (no pick up) are automatically entered each year. Edit as needed to add additional no shipping dates. Used for time in transit delivery date.', '6', '33', NULL, now(), NULL, NULL)");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Display Weight', 'MODULE_SHIPPING_UPSXML_WEIGHT', 'True', 'Do you want to show number of packages and package weight?', '6', '28', 'tep_cfg_select_option(array(\'True\', \'False\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, set_function, date_added) values ('Require Signature', 'MODULE_SHIPPING_UPSXML_REQSIG', 'No', 'Do you want to require customers to sign for packages over a certain value?', '6', '29', 'tep_cfg_select_option(array(\'No\', \'Package Value\', \'Shipment Value\'), ', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Signature Threshold', 'MODULE_SHIPPING_UPSXML_SIG_THRESHOLD', '1000', 'Require signature when value exceeds:', '6', '30', now())");
        tep_db_query("insert into " . TABLE_CONFIGURATION . " (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added) values ('Adult Signature Threshold', 'MODULE_SHIPPING_UPSXML_ADULTSIG_THRESHOLD', '1500', 'Require adult signature when value exceeds (overrides standard signature when values overlap):', '6', '30', now())");
    }

    //****************
    function remove() {
        tep_db_query("delete from " . TABLE_CONFIGURATION . " where configuration_key in ('" . implode("', '", $this->keys()) . "')");
    }

    //*************
    function keys() {
        return array('MODULE_SHIPPING_UPSXML_RATES_STATUS', 'MODULE_SHIPPING_UPSXML_RATES_ACCESS_KEY', 'MODULE_SHIPPING_UPSXML_RATES_USERNAME', 'MODULE_SHIPPING_UPSXML_RATES_PASSWORD', 'MODULE_SHIPPING_UPSXML_RATES_PICKUP_METHOD', 'MODULE_SHIPPING_UPSXML_RATES_PACKAGE_TYPE', 'MODULE_SHIPPING_UPSXML_RATES_CUSTOMER_CLASSIFICATION_CODE', 'MODULE_SHIPPING_UPSXML_RATES_ORIGIN', 'MODULE_SHIPPING_UPSXML_RATES_SHIPNAME', 'MODULE_SHIPPING_UPSXML_RATES_ADDRESS1', 'MODULE_SHIPPING_UPSXML_RATES_ADDRESS2', 'MODULE_SHIPPING_UPSXML_RATES_ADDRESS3', 'MODULE_SHIPPING_UPSXML_RATES_CITY', 'MODULE_SHIPPING_UPSXML_RATES_STATEPROV', 'MODULE_SHIPPING_UPSXML_RATES_COUNTRY', 'MODULE_SHIPPING_UPSXML_RATES_POSTALCODE', 'MODULE_SHIPPING_UPSXML_RATES_MODE', 'MODULE_SHIPPING_UPSXML_RATES_QUOTE_TYPE', 'MODULE_SHIPPING_UPSXML_RATES_USE_NEGOTIATED_RATES', 'MODULE_SHIPPING_UPSXML_RATES_UPS_ACCOUNT_NUMBER', 'MODULE_SHIPPING_UPSXML_RATES_MANUAL_NEGOTIATED_RATE', 'MODULE_SHIPPING_UPSXML_HANDLING_TYPE', 'MODULE_SHIPPING_UPSXML_HANDLING_CHARGED', 'MODULE_SHIPPING_UPSXML_RATES_HANDLING', 'MODULE_SHIPPING_UPSXML_INSURE', 'MODULE_SHIPPING_UPSXML_CURRENCY_CODE', 'MODULE_SHIPPING_UPSXML_RATES_TAX_CLASS', 'MODULE_SHIPPING_UPSXML_RATES_ZONE', 'MODULE_SHIPPING_UPSXML_RATES_SORT_ORDER', 'MODULE_SHIPPING_UPSXML_TYPES', 'MODULE_SHIPPING_UPSXML_SUREPOST', 'MODULE_SHIPPING_UPSXML_SUREPOST_MAXVAL', 'MODULE_SHIPPING_UPSXML_SHIPPING_DAYS_DELAY', 'MODULE_SHIPPING_UPSXML_EMAIL_ERRORS', 'MODULE_SHIPPING_UPSXML_RATES_TIME_IN_TRANSIT_VIEW', 'MODULE_SHIPPING_UPSXML_STORE_DAYS', 'MODULE_SHIPPING_UPSXML_UPS_PICKUP_DAYS', 'MODULE_SHIPPING_UPSXML_UPS_HOLIDAYS', 'MODULE_SHIPPING_UPSXML_WEIGHT', 'MODULE_SHIPPING_UPSXML_REQSIG', 'MODULE_SHIPPING_UPSXML_SIG_THRESHOLD', 'MODULE_SHIPPING_UPSXML_ADULTSIG_THRESHOLD');
    }

    //***********************
    function _upsProduct($prod){
        $this->_upsProductCode = $prod;
    }

    //**********************************************
    function _upsOrigin($city, $stateprov, $country, $postal){
        $this->_upsOriginCity = $city;
        $this->_upsOriginStateProv = $stateprov;
        $this->_upsOriginCountryCode = $country;
        $this->_upsOriginPostalCode = str_replace(array(' ', '-'), '', $postal);
        if ($this->_upsOriginCountryCode == 'US') {
          $territories = array('AS','FM','GU','MH','MP','PR','PW','VI');
          if (in_array($this->_upsOriginStateProv,$territories)) {
            $this->_upsOriginCountryCode = $stateprov;
          }
        }
    }

    //**********************************************
    function _upsDest($address) {
        // UPS purports that if the origin is left out, it defaults to the account's location. Yeah, right.
        $state = $address['state'];
        $zone_query = tep_db_query("select zone_code from " . TABLE_ZONES . " where zone_name = '" .  tep_db_input($state) . "' and zone_country_id = '" . tep_db_input($address['country']['id']) . "' and length(zone_code) > 0");
        if (tep_db_num_rows($zone_query)) {
          $zone = tep_db_fetch_array($zone_query);
          $state = $zone['zone_code'];
        }
        $this->_upsDestCity = $address['city'];
        $this->_upsDestStateProv = $state;
        $this->_upsDestCountryCode = $address['country']['iso_code_2'];
        $this->_upsDestPostalCode = str_replace(array(' ', '-'), '', $address['postcode']);
        if ($this->_upsDestCountryCode == 'US') {
          $territories = array('AS','FM','GU','MH','MP','PR','PW','VI');
          if (in_array($this->_upsDestStateProv,$territories)) {
            $this->_upsDestCountryCode = $stateprov;
          }
        }
				if (tep_not_null($address['company'])) {
					$this->_upsDestName = htmlspecialchars($address['company'], ENT_QUOTES);
				} else {
					$this->_upsDestName = htmlspecialchars($address['firstname'] . ' ' . $address['lastname'], ENT_QUOTES);
				}
				$this->_upsDestAddress1 = htmlspecialchars($address['street_address'], ENT_QUOTES);
				$this->_upsDestPObox = preg_match('/^P\.?\s*O\.?\s*BOX|^BOX\s*(#\s*)?\d+$|^P\.?\s*O\.?\s*(#\s*)?\d+$|^POST\s*OFFICE\s*/i', $this->_upsDestAddress1);
				$this->_upsDestAddress2 = $this->_upsDestAddress3 = '';
				if (isset($address['address2'])) { // extra address line contribution enabled
					if (tep_not_null($address['address2'])) $this->_upsDestAddress2 = htmlspecialchars($address['address2'], ENT_QUOTES);
					if (tep_not_null($address['suburb'])) $this->_upsDestAddress3 = htmlspecialchars($address['suburb'], ENT_QUOTES);
				} else { // standard osCommerce
					if (tep_not_null($address['suburb'])) $this->_upsDestAddress2 = htmlspecialchars($address['suburb'], ENT_QUOTES);
				}
    }

    //************************
    function _upsAction($action) {
        // rate - Single Quote; shop - All Available Quotes
        $this->_upsActionCode = $action;
    }

    //********************************************
    function _addItem($length, $width, $height, $weight, $price = 0 ) {
      // Add box or item to shipment list.
      $index = $this->items_qty;
      $this->item_length[$index] = ($length ? $length : 0 );
      $this->item_width[$index] = ($width ? $width : 0 );
      $this->item_height[$index] = ($height ? $height : 0 );
      $this->item_weight[$index] = ($weight == 0 ? 0.1 : $weight);
      $this->item_price[$index] = $price;
      $this->items_qty++;
    }

    //*********************
    function _upsGetQuote() {
        // Create the access request
        $accessRequestHeader =
        "<?xml version=\"1.0\"?>\n".
        "<AccessRequest xml:lang=\"en-US\">\n".
        "   <AccessLicenseNumber>". $this->access_key ."</AccessLicenseNumber>\n".
        "   <UserId>". $this->access_username ."</UserId>\n".
        "   <Password>". $this->access_password ."</Password>\n".
        "</AccessRequest>\n";

        $ratingServiceSelectionRequestHeader =
        "<?xml version=\"1.0\"?>\n".
        "<RatingServiceSelectionRequest xml:lang=\"en-US\">\n".
        "  <Request>\n".
        "    <TransactionReference>\n".
        "      <CustomerContext>Rating and Service</CustomerContext>\n".
        "      <XpciVersion>". $this->xpci_version ."</XpciVersion>\n".
        "    </TransactionReference>\n".
        "    <RequestAction>Rate</RequestAction>\n".
        "    <RequestOption>Shoptimeintransit</RequestOption>\n".
        "    <Subversion>" . $this->api_subversion . "</Subversion>\n".
        "  </Request>\n";
        // according to UPS the CustomerClassification and PickupType containers should
        // not be present when the origin country is non-US see:
        // http://forums.oscommerce.com/index.php?s=&showtopic=49382&view=findpost&p=730947
        if ($this->origin_country == 'US') {
        $ratingServiceSelectionRequestHeader .=
        "  <PickupType>\n".
        "    <Code>". $this->pickup_methods[$this->pickup_method] ."</Code>\n".
        "  </PickupType>\n" .
        "  <CustomerClassification>\n".
        "    <Code>". $this->customer_classification ."</Code>\n".
        "  </CustomerClassification>\n";
        }
        $ratingServiceSelectionRequestHeader .=
        "  <Shipment>\n".
        "    <Shipper>\n";
        if ($this->use_negotiated_rates == 'True') {
        $ratingServiceSelectionRequestHeader .=
        "      <ShipperNumber>" . $this->access_account_number . "</ShipperNumber>\n";
        }
        if (tep_not_null(MODULE_SHIPPING_UPSXML_RATES_SHIPNAME)) {
        $ratingServiceSelectionRequestHeader .=
        "      <Name>" . MODULE_SHIPPING_UPSXML_RATES_SHIPNAME . "</Name>\n";
        }
        $ratingServiceSelectionRequestHeader .=
        "      <Address>\n";
        if (tep_not_null(MODULE_SHIPPING_UPSXML_RATES_ADDRESS1)) {
        $ratingServiceSelectionRequestHeader .=
        "        <AddressLine1>" . MODULE_SHIPPING_UPSXML_RATES_ADDRESS1 . "</AddressLine1>\n";
        }
        if (tep_not_null(MODULE_SHIPPING_UPSXML_RATES_ADDRESS2)) {
        $ratingServiceSelectionRequestHeader .=
        "        <AddressLine2>" . MODULE_SHIPPING_UPSXML_RATES_ADDRESS2 . "</AddressLine2>\n";
        }
        if (tep_not_null(MODULE_SHIPPING_UPSXML_RATES_ADDRESS3)) {
        $ratingServiceSelectionRequestHeader .=
        "        <AddressLine3>" . MODULE_SHIPPING_UPSXML_RATES_ADDRESS3 . "</AddressLine3>\n";
        }
        $ratingServiceSelectionRequestHeader .=
        "        <City>". $this->_upsOriginCity ."</City>\n".
        "        <StateProvinceCode>". $this->_upsOriginStateProv ."</StateProvinceCode>\n".
        "        <CountryCode>". $this->_upsOriginCountryCode ."</CountryCode>\n".
        "        <PostalCode>". $this->_upsOriginPostalCode ."</PostalCode>\n".
        "      </Address>\n".
        "    </Shipper>\n".
        "    <ShipTo>\n".
        "      <CompanyName>" . $this->_upsDestName . "</CompanyName>\n" .
        "      <Address>\n".
        "        <AddressLine1>" . $this->_upsDestAddress1 . "</AddressLine1>\n" .
				($this->_upsDestPObox ? "        <POBoxIndicator/>\n" : '') .
        (tep_not_null($this->_upsDestAddress2) ? "        <AddressLine2>" . $this->_upsDestAddress2 . "</AddressLine2>\n" : "") .
        (tep_not_null($this->_upsDestAddress3) ? "        <AddressLine3>" . $this->_upsDestAddress3 . "</AddressLine3>\n" : "") .
        "        <City>". $this->_upsDestCity ."</City>\n".
        "        <StateProvinceCode>". $this->_upsDestStateProv ."</StateProvinceCode>\n".
        "        <CountryCode>". $this->_upsDestCountryCode ."</CountryCode>\n".
        "        <PostalCode>". $this->_upsDestPostalCode ."</PostalCode>\n".
        ($this->quote_type == "Residential" ? "        <ResidentialAddressIndicator/>\n" : "") .
        "      </Address>\n".
        "    </ShipTo>\n".
        "    <NumOfPieces>" . $this->NumPieces . "</NumOfPieces>\n" .
        "    <ShipmentTotalWeight>\n".
        "      <UnitOfMeasurement>\n".
        "        <Code>" . $this->unit_weight . "</Code>\n".
        "      </UnitOfMeasurement>\n".
        "      <Weight>" . ceil($this->totalWeight) . "</Weight>\n".
        "    </ShipmentTotalWeight>\n".
        "    <InvoiceLineTotal>\n".
        "      <CurrencyCode>" . MODULE_SHIPPING_UPSXML_CURRENCY_CODE . "</CurrencyCode>\n".
        "      <MonetaryValue>" . $this->pkgvalue . "</MonetaryValue>\n".
        "    </InvoiceLineTotal>\n".
        "    <DeliveryTimeInformation>\n".
        "      <PackageBillType>03</PackageBillType>\n".
        "      <Pickup>\n".
        "        <Date>" . $this->_set_ship_date() . "</Date>\n".
        "      </Pickup>\n".
        "    </DeliveryTimeInformation>\n";
        if ($this->sig_level == 'S') { // shipment level signature requirement
          $conf_type = 0;
          if ($this->all_adult_sig || ($this->pkgvalue >= MODULE_SHIPPING_UPSXML_ADULTSIG_THRESHOLD)) {
            $conf_type = 2;
						$this->sigreq = 'a';
          } elseif ($this->all_sig || ($this->pkgvalue >= MODULE_SHIPPING_UPSXML_SIG_THRESHOLD)) {
            $conf_type = 1;
						$this->sigreq = 's';
          }
          if ($conf_type > 0) {
            $ratingServiceSelectionRequestHeader .=
            "    <ShipmentServiceOptions>\n".
            "      <DeliveryConfirmation>\n".
            "        <DCISType>" . $conf_type . "</DCISType>\n".
            "      </DeliveryConfirmation>\n".
            "    </ShipmentServiceOptions>\n";
          }
        } // end if shipment level signature requirement
        $numgroups = ceil($this->items_qty / 50); // UPS can only process 50 packages at once
        $xmlResult = '';
        for ($g = 0; $g < $numgroups; $g++) { // process each group of packages
        $ratingServiceSelectionRequestPackageContent = '';
        $start = $g * 50;
        $end = ($g + 1 == $numgroups) ? $this->items_qty : $start + 50; // if last group end with number of packages otherwise do 50 more
        for ($i = $start; $i < $end; $i++) {

            $ratingServiceSelectionRequestPackageContent .=
            "       <Package>\n".
            "           <PackagingType>\n".
            "               <Code>". $this->package_types[$this->package_type] ."</Code>\n".
            "           </PackagingType>\n";
            if ($this->dimensions_support > 0 && ($this->item_length[$i] > 0 ) && ($this->item_width[$i] > 0 ) && ($this->item_height[$i] > 0)) {

                $ratingServiceSelectionRequestPackageContent .=
                "           <Dimensions>\n".
                "               <UnitOfMeasurement>\n".
                "                   <Code>". $this->unit_length ."</Code>\n".
                "               </UnitOfMeasurement>\n".
                "               <Length>". (float)$this->item_length[$i] ."</Length>\n".
                "               <Width>". (float)$this->item_width[$i] ."</Width>\n".
                "               <Height>". (float)$this->item_height[$i] ."</Height>\n".
                "           </Dimensions>\n";
            } // end if dimensions support

            $ratingServiceSelectionRequestPackageContent .=
            "           <PackageWeight>\n".
            "               <UnitOfMeasurement>\n".
            "                   <Code>". $this->unit_weight ."</Code>\n".
            "               </UnitOfMeasurement>\n".
            "               <Weight>". round($this->item_weight[$i], 1) ."</Weight>\n".
            "           </PackageWeight>\n" .
            "           <PackageServiceOptions>\n";
            if ($this->insure_package == true) {
            $ratingServiceSelectionRequestPackageContent .=
            "               <InsuredValue>\n" .
            "                   <CurrencyCode>" . MODULE_SHIPPING_UPSXML_CURRENCY_CODE . "</CurrencyCode>\n".
            "                   <MonetaryValue>" . $this->item_price[$i] . "</MonetaryValue>\n".
            "               </InsuredValue>\n";
            } // end if ($this->insure_package == true)
            if ($this->sig_level == 'P') { // package level signature requirement
              $conf_type = 0;
              if ($this->all_adult_sig || ($this->item_price[$i] >= MODULE_SHIPPING_UPSXML_ADULTSIG_THRESHOLD)) {
                $conf_type = 3;
								$this->sigreq = 'a';
              } elseif ($this->all_sig || ($this->item_price[$i] >= MODULE_SHIPPING_UPSXML_SIG_THRESHOLD)) {
                $conf_type = 2;
								if ($this->sigreq == '') $this->sigreq = 's';
              }
              if ($conf_type > 0) {
            $ratingServiceSelectionRequestPackageContent .=
            "               <DeliveryConfirmation>\n".
            "                   <DCISType>" . $conf_type . "</DCISType>\n".
            "               </DeliveryConfirmation>\n";
              }
            } // end if package level signature requirement
            $ratingServiceSelectionRequestPackageContent .=
            "           </PackageServiceOptions>\n" .
            "       </Package>\n";
        }

        $ratingServiceSelectionRequestFooter = '';
        if ($this->use_negotiated_rates == 'True') {
        $ratingServiceSelectionRequestFooter .=
            "    <RateInformation>\n".
            "      <NegotiatedRatesIndicator/>\n".
            "    </RateInformation>\n";
        }
        $ratingServiceSelectionRequestFooter .=
        "  </Shipment>\n" .
        "</RatingServiceSelectionRequest>\n";

        $xmlRequest = $accessRequestHeader .
        $ratingServiceSelectionRequestHeader .
        $ratingServiceSelectionRequestPackageContent .
        $ratingServiceSelectionRequestFooter;
				// debugging line
				//echo '<pre>'. htmlspecialchars($xmlRequest) . "</pre><br><br><br>\n";

        //post request $strXML;
        $result = $this->_post($this->protocol, $this->host, $this->port, $this->path, $this->version, $this->timeout, $xmlRequest);
        if ($xmlResult == '') { // if first group of packages
          $xmlResult = $result;
        } else { //if second group of packages then results must be combined into one larger result
          if (strpos($xmlResult, '</RatingServiceSelectionResponse>') !== false)
            $xmlResult = substr($xmlResult, 0, strpos($xmlResult, '</RatingServiceSelectionResponse>'));
          if (strpos($result, '</Response>') !== false)
            $result = substr($result, strpos($result, '</Response>') + 11);
          $xmlResult .= $result;
        }
				// debugging line
				//echo htmlspecialchars($xmlResult) . "<br><br><br>\n";
      } //end groups loop
			$services = $this->_parseResult($xmlResult);
			
			// SurePost is never returned by rate shopping even if the shipment qualifies and the shippers account is enabled to use SurePost
			// Check to see if this shipment qualifies for SurePost before attempting to retrieve a quote
			if ((MODULE_SHIPPING_UPSXML_SUREPOST == 'Yes') && ($this->_upsOriginCountryCode == 'US') && in_array($this->_upsDestCountryCode, array('US', 'AS','FM','GU','MH','MP','PR','PW','VI')) && ($this->NumPieces == 1) && ($this->totalWeight <= 70) && ($this->pkgvalue <= MODULE_SHIPPING_UPSXML_SUREPOST_MAXVAL) && (strlen($this->access_account_number) > 0)) {
				$SPrate = $this->_upsGetSurePostQuote();
				if (is_array($services) && is_array($SPrate)) {
					foreach ($SPrate as $code => $svc) $services[$code] = $svc; // array_merge screws up the codes so we use this instead
				} elseif ((!is_array($services) && is_array($SPrate)) || ((($services === false) || empty($services)) && is_string($SPrate))) {
					$services = $SPrate;
				}
			} elseif (!is_array($services) && $this->_upsDestPObox) {
				$services = MODULE_SHIPPING_UPSXML_ERROR_POBOX;
			}
      return $services;
    } // end _upsGetQuote

    //******************************************************************
    function _post($protocol, $host, $port, $path, $version, $timeout, $xmlRequest) {
        $url = $protocol."://".$host.":".$port.$path;
        if (isset($this->logfile)) {
            error_log("------------------------------------------\n", 3, $this->logfile);
            error_log("DATE AND TIME: ".date('Y-m-d H:i:s')."\n", 3, $this->logfile);
            error_log("UPS URL: " . $url . "\n", 3, $this->logfile);
        }
        if (function_exists('exec') && $this->use_exec == '1' ) {
            exec('which curl', $curl_output);
            if ($curl_output) {
                $curl_path = $curl_output[0];
            } else {
                $curl_path = 'curl'; // change this if necessary
            }
            if (isset($this->logfile)) {
                error_log("UPS REQUEST using exec(): " . $xmlRequest . "\n", 3, $this->logfile);
            }
            // add option -k to the statement: $command = "".$curl_path." -k -d \"". etcetera if you get
            // curl error 60: error setting certificate verify locations
            // using addslashes was the only way to avoid UPS returning the 1001 error: The XML document is not well formed
            $command = "".$curl_path." -d \"".addslashes($xmlRequest)."\" ".$url."";
            exec($command, $xmlResponse);
            if ( empty($xmlResponse) && isset($this->logfile)) { // using exec no curl errors can be retrieved
                error_log("Error from cURL using exec() since there is no \$xmlResponse\n", 3, $this->logfile);
            }
            if (isset($this->logfile)) {
                error_log("UPS RESPONSE using exec(): " . implode('', $xmlResponse) . "\n", 3, $this->logfile);
            }
        } elseif ($this->use_exec == '1') { // if NOT (function_exists('exec') && $this->use_exec == '1'
            if (isset($this->logfile)) {
                error_log("Sorry, exec() cannot be called\n", 3, $this->logfile);
            }
        } else { // default behavior: cURL is assumed to be compiled in PHP
            $ch = curl_init();
            curl_setopt($ch, CURLOPT_URL, $url);
            // uncomment the next line if you get curl error 60: error setting certificate verify locations
            // curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
            // uncommenting the next line is most likely not necessary in case of error 60
            // curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
            curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
            curl_setopt($ch, CURLOPT_HEADER, 0);
            curl_setopt($ch, CURLOPT_POST, 1);
            curl_setopt($ch, CURLOPT_POSTFIELDS, $xmlRequest);
            curl_setopt($ch, CURLOPT_TIMEOUT, (int)$timeout);

            if (isset($this->logfile)) {
                error_log("UPS REQUEST: " . $xmlRequest . "\n", 3, $this->logfile);
            }
            $xmlResponse = curl_exec($ch);
						if (curl_errno($ch) == 6) { // failed to resolve host on first try, make one more attempt before giving up
							$xmlResponse = curl_exec($ch);
						}
						if (curl_errno($ch)) {
							$err = array('code' => curl_errno($ch), 'desc' => curl_error($ch));
						}
            if (curl_errno($ch) && isset($this->logfile)) {
                $error_from_curl = sprintf('Error [%d]: %s', curl_errno($ch), curl_error($ch));
                error_log("Error from cURL: " . $error_from_curl . "\n", 3, $this->logfile);
            }
            // send email if enabled in the admin section
            if (curl_errno($ch) && $this->email_errors) {
                $error_from_curl = sprintf('Error [%d]: %s', curl_errno($ch), curl_error($ch));
                error_log("Error from cURL: " . $error_from_curl . " experienced by customer with id " . $_SESSION['customer_id'] . " " . $this->_customer_name . " on " . date('Y-m-d H:i:s'), 1, STORE_OWNER_EMAIL_ADDRESS);
            }
            // log errors to file ups_error.log when set
            if (curl_errno($ch) && isset($this->ups_error_file)) {
                $error_from_curl = sprintf('Error [%d]: %s', curl_errno($ch), curl_error($ch));
                error_log(date('Y-m-d H:i:s')."\tcURL\t" . $error_from_curl . "\t" . $_SESSION['customer_id'] . " " . $this->_customer_name."\n", 3, $this->ups_error_file);    
            }
            if (isset($this->logfile)) {
                error_log("UPS RESPONSE: " . $xmlResponse . "\n", 3, $this->logfile);
            }
            curl_close ($ch);
        }

        if(!$xmlResponse || strstr(strtolower(substr($xmlResponse, 0, 120)), "bad request"))  {
            /* Sometimes the UPS server responds with an HTML message (differing depending on whether the test server
               or the production server is used) but both have in the title tag "Bad request".
               Parsing this response will result in a fatal error:
               Call to a member function on a non-object in /blabla/includes/classes/xmldocument.php on line 57
               It only results in not showing Estimated Delivery Dates to the customer so avoiding the fatal error should do.
            */
						if (!isset($err)) $err = array('code' => '*', 'desc' => MODULE_SHIPPING_UPSXML_RATES_TEXT_COMM_UNKNOWN_ERROR);
            $xmlResponse = "<?xml version=\"1.0\"?>\n".
            "<RatingServiceSelectionResponse>\n".
            "   <Response>\n".
            "       <TransactionReference>\n".
            "           <CustomerContext>Rating and Service</CustomerContext>\n".
            "           <XpciVersion>". $this->xpci_version ."</XpciVersion>\n".
            "       </TransactionReference>\n".
            "       <ResponseStatusCode>0</ResponseStatusCode>\n".
            "       <ResponseStatusDescription>Connection Error</ResponseStatusDescription>\n".
						"       <Error>\n".
						"         <ErrorCode>" . $err['code'] . "</ErrorCode>\n".
						"         <ErrorDescription>" . MODULE_SHIPPING_UPSXML_RATES_TEXT_COMM_ERROR . '<br>' . $err['desc'] . "</ErrorDescription>\n".
						"       </Error>\n".
            "   </Response>\n".
            "</RatingServiceSelectionResponse>\n";
            return $xmlResponse;
        }
        if ($this->use_exec == '1') {
            return implode('', $xmlResponse); // $xmlResponse is an array in this case
        } else {
            return $xmlResponse;
        }
    }

    //*****************************
    function _parseResult($xmlResult) {
        // Parse XML message returned by the UPS post server.
        $doc = XML_unserialize ($xmlResult);
				
				// debug line
				//echo '<pre>'; print_r($doc); echo "</pre>";

        // Get version. Must be xpci version 1.0001 or this might not work.
        $responseVersion = $doc['RatingServiceSelectionResponse']['Response']['TransactionReference']['XpciVersion'];
        if ($this->xpci_version != $responseVersion) {
            $message = MODULE_SHIPPING_UPSXML_RATES_TEXT_COMM_VERSION_ERROR;
            return $message;
        }
        // Get response code: 1 = SUCCESS, 0 = FAIL
        $responseStatusCode = $doc['RatingServiceSelectionResponse']['Response']['ResponseStatusCode'];
        if ($responseStatusCode != '1') {
					if (isset($doc['RatingServiceSelectionResponse']['Response']['Error'][0])) {
						$errorMsg = '';
						foreach ($doc['RatingServiceSelectionResponse']['Response']['Error'] as $e) {
						  $errorMsg .= MODULE_SHIPPING_UPSXML_RATES_TEXT_ERROR . $e['ErrorCode'];
              $errorMsg .= ": ";
              $errorMsg .= $e['ErrorDescription'] . ' & ';
						}
						$errorMsg = rtrim($errorMsg, ' & ');
					} else {
            $errorMsg = MODULE_SHIPPING_UPSXML_RATES_TEXT_ERROR . $doc['RatingServiceSelectionResponse']['Response']['Error']['ErrorCode'];
            $errorMsg .= ": ";
            $errorMsg .= $doc['RatingServiceSelectionResponse']['Response']['Error']['ErrorDescription'];
					}
            // send email if enabled in the admin section
            if ($this->email_errors) {
                error_log("UPSXML Rates Error: " . $errorMsg . " experienced by customer with id " . $_SESSION['customer_id'] . " " . $this->_customer_name . " on " . date('Y-m-d H:i:s'), 1, STORE_OWNER_EMAIL_ADDRESS);
            }
            // log errors to file ups_error.log when set
            if (isset($this->ups_error_file)) {
                error_log(date('Y-m-d H:i:s')."\tRates\t" . $errorMsg . "\t" . $_SESSION['customer_id'] . " " . $this->_customer_name."\n", 3, $this->ups_error_file);    
            }
            return $errorMsg;
        }

        $ratedShipments = $doc['RatingServiceSelectionResponse']['RatedShipment'];
        $services = $upstemp = array();
        if (isset($doc['RatingServiceSelectionResponse']['RatedShipment'][0])) { // more than 1 rate
          for ($i = 0; $i < count($ratedShipments); $i++) {
            $serviceCode = $ratedShipments[$i]['Service']['Code'];
						if (in_array($serviceCode, $this->disallowed_methods)) continue;
						$deliveryday = $ratedShipments[$i]['TimeInTransit']['ServiceSummary']['EstimatedArrival']['DayOfWeek'];
            $serviceMethod = constant('UPS_SERVICE_CODE_' . $serviceCode);
						if ($deliveryday == 'SAT') {
							$serviceMethod .= UPS_SATDEL;
						  $serviceCode .= 'S'; // add to service code since the same code will be returned for the same method without Saturday delivery
						}
            if ($this->use_negotiated_rates == 'True' && isset($ratedShipments[$i]['NegotiatedRates']['NetSummaryCharges']['GrandTotal']['MonetaryValue'])) {
              $totalCharge = $ratedShipments[$i]['NegotiatedRates']['NetSummaryCharges']['GrandTotal']['MonetaryValue'];
            } elseif ($this->manual_negotiated_rate > 0) {
              $totalCharge = $ratedShipments[$i]['TotalCharges']['MonetaryValue'] * ($this->manual_negotiated_rate/100);
            } else {
// standard UPS rates
              $totalCharge = $ratedShipments[$i]['TotalCharges']['MonetaryValue'];
            }
            if (!($serviceCode && $totalCharge)) {
              continue;
            } 
            $ratedPackages = $ratedShipments[$i]['RatedPackage']; // only do this once for the first service given
            if (isset($ratedShipments[$i]['RatedPackage'][0])) { // multidimensional array of packages
              $boxCount = count($ratedPackages);
            } else {
              $boxCount = 1; // if there is only one package count($ratedPackages) returns
              // the number of fields in the array like TransportationCharges and BillingWeight
            }
            // if more than one group of packages, service codes will be repeated and therefore data needs to be combined
						if (!isset($upstemp[$serviceCode]['charge'])) $upstemp[$serviceCode]['charge'] = 0;
						if (!isset($upstemp[$serviceCode]['boxes'])) $upstemp[$serviceCode]['boxes'] = 0;
						if (!isset($upstemp[$serviceCode]['billed_weight'])) $upstemp[$serviceCode]['billed_weight'] = 0;
            $upstemp[$serviceCode]['method'] = $serviceMethod;
            $upstemp[$serviceCode]['edd'] = $this->_expand_date($ratedShipments[$i]['TimeInTransit']['ServiceSummary']['EstimatedArrival']['Arrival']['Date']);
            $upstemp[$serviceCode]['charge'] += $totalCharge;
            $upstemp[$serviceCode]['boxes'] += $boxCount;
            $upstemp[$serviceCode]['billed_weight'] += $ratedShipments[$i]['BillingWeight']['Weight'];
            $upstemp[$serviceCode]['weight_code'] = $ratedShipments[$i]['BillingWeight']['UnitOfMeasurement']['Code'];
          } // end for ($i = 0; $i < count($ratedShipments); $i++)
          foreach ($upstemp as $code => $svc) {
            if ($svc['boxes'] == $this->NumPieces) $services[$code] = $svc; // only return methods that account for all boxes to be shipped
          }
        } elseif (isset($doc['RatingServiceSelectionResponse']['RatedShipment'])) { // only 1 rate
          $serviceCode = $ratedShipments['Service']['Code'];
					if (in_array($serviceCode, $this->disallowed_methods)) return false;
					if (in_array($serviceCode, array('92', '93', '94', '95'))) { // SurePost cannot return time in transit information
					  $deliveryday = 'UNK';
					} else {
					  $deliveryday = $ratedShipments['TimeInTransit']['ServiceSummary']['EstimatedArrival']['DayOfWeek'];
					}
          $serviceMethod = constant('UPS_SERVICE_CODE_' . $serviceCode);
  				if ($deliveryday == 'SAT') {
						$serviceMethod .= UPS_SATDEL;
					  $serviceCode .= 'S'; // add to service code since the same code will be returned for the same method without Saturday delivery
					}
          if ($this->use_negotiated_rates == 'True' && isset($ratedShipments['NegotiatedRates']['NetSummaryCharges']['GrandTotal']['MonetaryValue'])) {
            $totalCharge = $ratedShipments['NegotiatedRates']['NetSummaryCharges']['GrandTotal']['MonetaryValue'];
          } elseif ($this->manual_negotiated_rate > 0) {
            $totalCharge = $ratedShipments['TotalCharges']['MonetaryValue'] * ($this->manual_negotiated_rate/100);
          } else { // standard UPS rates
            $totalCharge = $ratedShipments['TotalCharges']['MonetaryValue'];
          }
          if (!($serviceCode && $totalCharge)) {
            return false; // is false
          } 
          $ratedPackages = $ratedShipments['RatedPackage']; // only do this once for the first service given
          if (isset($ratedShipments['RatedPackage'][0])) { // multidimensional array of packages
            $boxCount = count($ratedPackages);
          } else {
            $boxCount = 1; // if there is only one package count($ratedPackages) returns
            // the number of fields in the array like TransportationCharges and BillingWeight
          }
          $upstemp[$serviceCode]['charge'] = $totalCharge;
          $upstemp[$serviceCode]['method'] = $serviceMethod;
					if (in_array($serviceCode, array('92', '93', '94', '95'))) { // SurePost cannot return time in transit information
					  $upstemp[$serviceCode]['edd'] = MODULE_SHIPPING_UPSXML_TEXT_SUREPOST_DEL;
					} else {
            $upstemp[$serviceCode]['edd'] = $this->_expand_date($ratedShipments['TimeInTransit']['ServiceSummary']['EstimatedArrival']['Arrival']['Date']);
					}
          $upstemp[$serviceCode]['boxes'] = $boxCount;
          $upstemp[$serviceCode]['billed_weight'] = $ratedShipments['BillingWeight']['Weight'];
          $upstemp[$serviceCode]['weight_code'] = $ratedShipments['BillingWeight']['UnitOfMeasurement']['Code'];
					if ($boxCount == $this->NumPieces) $services = $upstemp; // only return methods that account for all boxes to be shipped
        }
        return $services;
    }

    function _upsGetSurePostQuote() {
        // Create the access request
        $accessRequestHeader =
        "<?xml version=\"1.0\"?>\n".
        "<AccessRequest xml:lang=\"en-US\">\n".
        "   <AccessLicenseNumber>". $this->access_key ."</AccessLicenseNumber>\n".
        "   <UserId>". $this->access_username ."</UserId>\n".
        "   <Password>". $this->access_password ."</Password>\n".
        "</AccessRequest>\n";
        $ratingServiceSelectionRequestHeader =
        "<?xml version=\"1.0\"?>\n".
        "<RatingServiceSelectionRequest xml:lang=\"en-US\">\n".
        "  <Request>\n".
        "    <TransactionReference>\n".
        "      <CustomerContext>Rating and Service</CustomerContext>\n".
        "      <XpciVersion>". $this->xpci_version ."</XpciVersion>\n".
        "    </TransactionReference>\n".
        "    <RequestAction>Rate</RequestAction>\n".
        "    <RequestOption>Rate</RequestOption>\n".
        "    <Subversion>" . $this->api_subversion . "</Subversion>\n".
        "  </Request>\n".
        "  <PickupType>\n".
        "    <Code>". $this->pickup_methods[$this->pickup_method] ."</Code>\n".
        "  </PickupType>\n" .
        "  <CustomerClassification>\n".
        "    <Code>". $this->customer_classification ."</Code>\n".
        "  </CustomerClassification>\n".
        "  <Shipment>\n".
        "    <Service>\n".
        "      <Code>". ($this->item_weight[0] < 1 ? '92' : '93') ."</Code>\n".
        "    </Service>\n".
        "    <Shipper>\n".
        "      <ShipperNumber>" . $this->access_account_number . "</ShipperNumber>\n";
        if (tep_not_null(MODULE_SHIPPING_UPSXML_RATES_SHIPNAME)) {
        $ratingServiceSelectionRequestHeader .=
        "      <Name>" . MODULE_SHIPPING_UPSXML_RATES_SHIPNAME . "</Name>\n";
        }
        $ratingServiceSelectionRequestHeader .=
        "      <Address>\n";
        if (tep_not_null(MODULE_SHIPPING_UPSXML_RATES_ADDRESS1)) {
        $ratingServiceSelectionRequestHeader .=
        "        <AddressLine1>" . MODULE_SHIPPING_UPSXML_RATES_ADDRESS1 . "</AddressLine1>\n";
        }
        if (tep_not_null(MODULE_SHIPPING_UPSXML_RATES_ADDRESS2)) {
        $ratingServiceSelectionRequestHeader .=
        "        <AddressLine2>" . MODULE_SHIPPING_UPSXML_RATES_ADDRESS2 . "</AddressLine2>\n";
        }
        if (tep_not_null(MODULE_SHIPPING_UPSXML_RATES_ADDRESS3)) {
        $ratingServiceSelectionRequestHeader .=
        "        <AddressLine3>" . MODULE_SHIPPING_UPSXML_RATES_ADDRESS3 . "</AddressLine3>\n";
        }
        $ratingServiceSelectionRequestHeader .=
        "        <City>". $this->_upsOriginCity ."</City>\n".
        "        <StateProvinceCode>". $this->_upsOriginStateProv ."</StateProvinceCode>\n".
        "        <CountryCode>". $this->_upsOriginCountryCode ."</CountryCode>\n".
        "        <PostalCode>". $this->_upsOriginPostalCode ."</PostalCode>\n".
        "      </Address>\n".
        "    </Shipper>\n".
        "    <ShipTo>\n".
        "      <CompanyName>" . $this->_upsDestName . "</CompanyName>\n" .
        "      <Address>\n".
        "        <AddressLine1>" . $this->_upsDestAddress1 . "</AddressLine1>\n" .
				($this->_upsDestPObox ? "        <POBoxIndicator/>\n" : '') .
        (tep_not_null($this->_upsDestAddress2) ? "        <AddressLine2>" . $this->_upsDestAddress2 . "</AddressLine2>\n" : "") .
        (tep_not_null($this->_upsDestAddress3) ? "        <AddressLine3>" . $this->_upsDestAddress3 . "</AddressLine3>\n" : "") .
        "        <City>". $this->_upsDestCity ."</City>\n".
        "        <StateProvinceCode>". $this->_upsDestStateProv ."</StateProvinceCode>\n".
        "        <CountryCode>". $this->_upsDestCountryCode ."</CountryCode>\n".
        "        <PostalCode>". $this->_upsDestPostalCode ."</PostalCode>\n".
        ($this->quote_type == "Residential" ? "        <ResidentialAddressIndicator/>\n" : "") .
        "      </Address>\n".
        "    </ShipTo>\n".
        "    <NumOfPieces>" . $this->NumPieces . "</NumOfPieces>\n" .
        "    <ShipmentTotalWeight>\n".
        "      <UnitOfMeasurement>\n".
        "        <Code>" . $this->unit_weight . "</Code>\n".
        "      </UnitOfMeasurement>\n".
        "      <Weight>" . $this->totalWeight . "</Weight>\n".
        "    </ShipmentTotalWeight>\n".
        "    <InvoiceLineTotal>\n".
        "      <CurrencyCode>" . MODULE_SHIPPING_UPSXML_CURRENCY_CODE . "</CurrencyCode>\n".
        "      <MonetaryValue>" . $this->pkgvalue . "</MonetaryValue>\n".
        "    </InvoiceLineTotal>\n".
        "    <DeliveryTimeInformation>\n".
        "      <PackageBillType>03</PackageBillType>\n".
        "      <Pickup>\n".
        "        <Date>" . $this->_set_ship_date() . "</Date>\n".
        "      </Pickup>\n".
        "    </DeliveryTimeInformation>\n";
        $ratingServiceSelectionRequestPackageContent =
        "    <Package>\n".
        "      <PackagingType>\n".
        "        <Code>". $this->package_types[$this->package_type] ."</Code>\n".
        "      </PackagingType>\n";
        if ($this->dimensions_support > 0 && ($this->item_length[0] > 0 ) && ($this->item_width[0] > 0 ) && ($this->item_height[0] > 0)) {
          $ratingServiceSelectionRequestPackageContent .=
          "      <Dimensions>\n".
          "        <UnitOfMeasurement>\n".
          "          <Code>". $this->unit_length ."</Code>\n".
          "        </UnitOfMeasurement>\n".
          "        <Length>". (float)$this->item_length[0] ."</Length>\n".
          "        <Width>". (float)$this->item_width[0] ."</Width>\n".
          "        <Height>". (float)$this->item_height[0] ."</Height>\n".
          "      </Dimensions>\n";
        } // end if dimensions support
        $ratingServiceSelectionRequestPackageContent .=
        "      <PackageWeight>\n".
        "        <UnitOfMeasurement>\n".
        "          <Code>". ($this->item_weight[0] < 1 ? 'OZS' : $this->unit_weight) ."</Code>\n".
        "        </UnitOfMeasurement>\n".
        "        <Weight>". ($this->item_weight[0] < 1 ? (16 * $this->item_weight[0]) : round($this->item_weight[0], 1)) ."</Weight>\n".
        "      </PackageWeight>\n" .
        "      <PackageServiceOptions>\n".
        "        <NonMachineableIndicator/>\n" .
        "      </PackageServiceOptions>\n" .
        "    </Package>\n";
        $ratingServiceSelectionRequestFooter = '';
        if ($this->use_negotiated_rates == 'True') {
        $ratingServiceSelectionRequestFooter .=
            "    <RateInformation>\n".
            "      <NegotiatedRatesIndicator/>\n".
            "    </RateInformation>\n";
        }
        $ratingServiceSelectionRequestFooter .=
        "  </Shipment>\n" .
        "</RatingServiceSelectionRequest>\n";

        $xmlRequest = $accessRequestHeader .
        $ratingServiceSelectionRequestHeader .
        $ratingServiceSelectionRequestPackageContent .
        $ratingServiceSelectionRequestFooter;
				// debugging line
				//echo '<pre>'. htmlspecialchars($xmlRequest) . "</pre><br><br><br>\n";

        //post request $strXML;
        $xmlResult = $this->_post($this->protocol, $this->host, $this->port, $this->path, $this->version, $this->timeout, $xmlRequest);
				// debugging line
				//echo htmlspecialchars($xmlResult) . "<br><br><br>\n";
      return $this->_parseResult($xmlResult);
    } // end _upsGetSurePostQuote


    // for Time In Transit
    function _set_ship_date() {
			// add delay to current date, subtract 1 from day of week since it returns 1-7 and array is indexed 0-6
      $sd = tep_db_fetch_array(tep_db_query("select date_add(curdate(), interval " . (int)MODULE_SHIPPING_UPSXML_SHIPPING_DAYS_DELAY . " day) as sd, dayofweek(date_add(curdate(), interval " . (int)MODULE_SHIPPING_UPSXML_SHIPPING_DAYS_DELAY . " day))-1 as dow"));
      // advance date until it falls on a day of the week that the store processes orders
      while (!in_array($this->dayname[$sd['dow']], $this->store_days)) $sd = tep_db_fetch_array(tep_db_query("select date_add('" . $sd['sd'] . "', interval 1 day) as sd, dayofweek(date_add('" . $sd['sd'] . "', interval 1 day))-1 as dow"));
      // advance date until it falls on a day of the week that UPS picks up packages
      while (!in_array($this->dayname[$sd['dow']], $this->ups_days) || in_array(substr(str_replace('-', '', $sd['sd']),4), $this->ups_holidays)) $sd = tep_db_fetch_array(tep_db_query("select date_add('" . $sd['sd'] . "', interval 1 day) as sd, dayofweek(date_add('" . $sd['sd'] . "', interval 1 day))-1 as dow"));
			$this->_ship_date = $sd['sd'];
      return str_replace('-', '', $sd['sd']);
		}
		
		function _expand_date($d) { // converts UPS format date of YYYYMMDD to MySQL format YYYY-MM-DD
			return substr($d, 0, 4) . '-' . substr($d, 4, 2) . '-' . substr($d, 6);
		}
		
// Next function used for sorting the shipping quotes on rate: low to high is default.
  function rate_sort_func ($a, $b) {
    
   $av = array_values($a);
   $av = $av[0];
   $bv = array_values($b);
   $bv = $bv[0];

//  return ($av == $bv) ? 0 : (($av < $bv) ? 1 : -1); // for having the high rates first
  return ($av == $bv) ? 0 : (($av > $bv) ? 1 : -1); // low rates first
  
  }
  
  // sets known days that UPS does not pick up this year
  function _update_ups_holidays($this_year) {
    // Christmas Eve, Christmas, New Years Eve, New Years Day, Independence Day
    $holidays = array('1224', '1225', '1231', '0101', '0704');
    // calculate Thanksgiving
    $tmp = mktime(1, 1, 1, 11, 1, $this_year);
    while (date('w', $tmp) != 4) $tmp += 86400; // find first Thursday of November
    $day = date('d', $tmp) + 21; // fourth Thursday of November, Thanksgiving
    $holidays[] = '11' . $day;
    $day++;
    $holidays[] = '11' . $day; // day after Thanksgiving
    // calculate Memorial Day
    $tmp = mktime(1, 1, 1, 5, 31, $this_year);
    while (date('w', $tmp) != 1) $tmp -= 86400;
    $day = date('d', $tmp); // last Monday of May, Memorial Day
    $holidays[] = '05' . $day;
    // calculate Labor Day
    $tmp = mktime(1, 1, 1, 9, 1, $this_year);
    while (date('w', $tmp) != 1) $tmp += 86400;
    $day = date('d', $tmp); // first Monday of September, Labor Day
    $holidays[] = '09'. $day;
    $this->ups_holidays = $holidays;
    tep_db_query("update " . TABLE_CONFIGURATION . " set configuration_value = '" . tep_db_input($this_year . ':' . implode(',', $holidays)) . "' where configuration_key = 'MODULE_SHIPPING_UPSXML_UPS_HOLIDAYS'");
  }
} // end class upsxml

// Next two functions are used only in the admin for disallowed shipping options.
// The (short) code numbers are stored in the database
// to stay below 255 characters. The defines themselves are found in the upsxml
// language file prefixed with UPS_SERVICE_CODE_ to avoid collisions with other shipping modules.
// They can be moved to admin/includes/function/general.php if you like but don't forget
// to remove them from this file in future updates or you will get an error in the admin
// about re-declaring functions
  function get_multioption_upsxml($values) {
    if (tep_not_null($values)) {
      $values_array = explode(',', $values);
      foreach ($values_array as $key => $_method) {
        if ($_method == '--none--') {
          $method = $_method;
        } else {
          $method = constant('UPS_SERVICE_CODE_' . trim($_method));
        }
        $readable_values_array[] = $method;
      }
      $readable_values = implode(', ', $readable_values_array);
      return $readable_values;
    } else {
      return '';
    }
  }
  
  function upsxml_cfg_select_multioption_indexed($select_array, $key_value, $key = '') {
    $key_values = explode( ", ", $key_value);
		
    $string = '';
    for ($i=0; $i<sizeof($select_array); $i++) {
      $name = ((tep_not_null($key)) ? 'configuration[' . $key . '][]' : 'configuration_value');
      $string .= '<br /><input type="checkbox" name="' . $name . '" value="' . $select_array[$i] . '"';
      if ( in_array($select_array[$i], $key_values) ) $string .= ' CHECKED';
      $string .= '> ' . constant('UPS_SERVICE_CODE_' . trim($select_array[$i]));
    } 
    $string .= '<input type="hidden" name="' . $name . '" value="--none--">';
    return $string;
  }
// Next function is used only in the admin for shipping days.
// It can be moved to admin/includes/function/general.php if you like but don't forget
// to remove it from this file in future updates or you will get an error in the admin
// about re-declaring functions
  function upsxml_cfg_select_multioption($select_array, $key_value, $key = '') {
    $string = '';
    $key_values = explode( ", ", $key_value);
    for ($i=0, $n=sizeof($select_array); $i<$n; $i++) {
      $name = ((tep_not_null($key)) ? 'configuration[' . $key . '][]' : 'configuration_value');
      $string .= '<br /><input type="checkbox" name="' . $name . '" value="' . $select_array[$i] . '"';
      if (in_array($select_array[$i], $key_values)) $string .= ' CHECKED';
      $string .= '> ' . $select_array[$i];
    }
    return $string;
  }
?>